package de.korap.services;

import java.io.IOException;
import java.io.PrintStream;
import java.util.Properties;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;

import de.korap.json.JsonTraverse;
import de.korap.json.TraversedPath;
import de.korap.services.glemm.client.GlemmClient;
import de.korap.services.utils.Utils;
import de.korap.services.LemmaResponse;

public class WordformRewrite {


	/* 
	 * traverse_and_rewrite_KoralQuery:
	 * Parameters:
	 * query     : JSON KoralQuery.
	 * props	 : wformServices.properties.
	 * 
	 * Notes:
	 * - 
	 * Returns	: 0 = rewrite done. 1 = no rewrite needed. > 1: error!
	 * - rewriten new Query if rewritten done.
	 * - null + no exception: no rewriting needed.
	 * - null + exception   : error.
	 * 04.03.20/FB 27.04.20/FB 
	 */
		 
	public static String traverse_and_rewrite_KoralQuery(
			String 		query, 
			Properties 	props, 
			PrintStream	fout,
			PrintStream ferr )
		
					throws IOException
	
	{
	final String
		func = "traverse_and_rewrite_KoralQuery";
	ObjectMapper
		objMapper = null;
	JsonNode
		jsonNode = null;
	TraversedPath
		tPath 	= null;
	LemmaResponse
		lemResp	= null;
	final boolean
		bDebug = true;
	int
		mode = JsonTraverse.REWRITE_EXP; // rewrite lemma query by expansion list.
	
	fout.printf("Debug: %s: ready for traversing query...\n", func);
	
	objMapper = new ObjectMapper();
	
	// get full query as a JsonNode tree:
	try {
		jsonNode = objMapper.readTree(query);
		
		// traverse query and collect lemma subqueries in a TraversedPath object:
		tPath = new TraversedPath(objMapper);
		JsonTraverse.traverseJsonTree(jsonNode, tPath, JsonTraverse.COLLECT, fout, ferr);
		} 
	catch (IOException e) {
		e.printStackTrace();
		throw e;
		}
	
	fout.printf("Debug: %s: query traversed.\n", func);
	
	// Exit if no lemma subqueries found:
	if( tPath == null || tPath.getLemmaCount() == 0 )
		{
		fout.printf("Info: %s: no lemma subquery found -> no rewrite needed.\n", func);
		return null; // no rewrite.
		}

	/* - - - Call GlemmServices to resolve lemma-Subqueries - - - 
	 * 
	 * This returns for each lemma a reference to a list of wordform instantiations
	 * (= instantiation list).
	 */ 
	
	GlemmClient
		glemmClient = new GlemmClient(props, fout, ferr);
	
	for(int i=0; i<tPath.getLemmaCount(); i++)
		{
		lemResp = glemmClient.sendQuery(tPath.getLemma(i));
		tPath.addLemmaResponse(lemResp);
		
		fout.printf("Debug: %s: lemma='%s': %d wfs found: %s.\n", func, 
				tPath.getLemma(i),
				lemResp.head_nWfs,
				lemResp.printPartialList());
		}

	// TODO glemmClient.close ?
	
	/*
	 *  - - - Rewriting KoralQuery with lists of wordforms:
	 *  
	 * - Rewrite KoralQuery: use either JsonTraverse.REWRITE or .REWRITE_EXP:
	 * - REWRITE_EXP = rewrite with expansion lists.
	 */
	
	JsonTraverse.traverseJsonTree(jsonNode, tPath, JsonTraverse.REWRITE_EXP, fout, ferr); 
	
	// OK: return newQuery formated in pretty json:
	fout.printf("Debug: %s: returning OK with jsonNode as a String.\n", func);
	
	return objMapper.writerWithDefaultPrettyPrinter().writeValueAsString(jsonNode);
		
	//System.out.printf("Debug: %s: returning OK and null String.\n", func);
	
	//return null; // OK

	} // traverse_and_rewrite_KoralQuery



} // class WordformRewrite

