package de.korap.services;

import javax.servlet.ServletContext;
import javax.ws.rs.Consumes;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.nio.charset.Charset;
import java.time.LocalDateTime;
import java.util.Locale;
import java.util.Properties;

import de.korap.services.WordformRewrite;
import de.korap.services.utils.Utils;


/*
 * v-0.1: 29.06.20/FB: 1st version.
 * v-0.1: 28.10.21/FB: 
 * 		- writes request time to log;
 *		- resets fnameOut/Err after call to GlemmServices. 
 * Notes:
 * - When submitting a KoralQuery [UTF-8] to the WformServices (if redirected
 *   to the GlemmServices, only chars that may be translated to ISO8859-1 (=Lat1)
 *   will be processed correctly. Glemm behind the scene can only treat
 *   German words, so this restriction is OK.
 */

@Path("/wformServices")

public class WordformServices {
	
	final private static String versionDate			= "28.10.21";
	final private static String version     		= "0.2";
	final private String 		fnameProps			= "/WEB-INF/WformServices.properties";
	      private static String GS_WorkingPath 		= "/home/bodmer/KorAP/GlemmServices/Tests"; // linux
	      private static String fnameOut 			= GS_WorkingPath + "/wformServices-t.log";
	      private static String fnameErr	 		= GS_WorkingPath + "/wformServices-t.err";
	final private static String filename_pluginHTML	= "WEB-INF/plugin.html";

	@Context 
	private ServletContext servletContext;
	
	Properties
		props = null; // properties in WEB-INF/WformServices.properties
	
	public static PrintStream fout	= null;
	public static PrintStream ferr	= null;		
		
	/*
	 *  log_Request:
	 *  
	 *  - write info about API Request.
	 *  Parameters:
	 *  requestPath	: path of webapp addressed by the request.
	 *  func		: name of the function which implements the request.
	 *  
	 *  19.10.21/FB
	 */
	
	private void log_Request(String requestPath, String func)
	
	{
	// transform "dateTtime" -> "date time":
	
	fout.printf("###\n");
	fout.printf("### %s: %s : vers='%s', vdate='%s' request='%s'.\n", func, requestPath, version, versionDate, 
						LocalDateTime.now().toString().replace("T",  " "));
	fout.printf("###\n");
	} // log_Request
		
	/* 
 	 * 
	 * - after loadProperties, sets the class variable with loaded property values.
	 * 
	 * 15.10.21/FB
	 */
	
	private void set_ConstantsFromProperties(Properties props)
	
	{
	GS_WorkingPath 	= Utils.getConstantFromProperties(props, "GS_WorkingPath", "/home/bodmer/KorAP/GlemmServices/Tests");
	fnameOut		= Utils.getConstantFromProperties(props, "fnameOut", 	   "WformServices-t.log");
	fnameOut		= GS_WorkingPath + "/" + fnameOut;
	fnameErr		= Utils.getConstantFromProperties(props, "fnameErr", 	   "WformServices-t.err");
	fnameErr		= GS_WorkingPath + "/" + fnameErr;
	
	} // set_ConstantsFromProperties
	
	/* init_and_log_Request
	 * 
	 * - load GlemmServices.properties, sets GlemmServices constants
	 *   and logs the request.
	 * - load properties and set constants before logging anything, because
	 *   logging will be done into the last stdout/stderr opened in glassfish's JVM,
	 *   e.g. in the log of another Service.
	 *    
	 * 19.10.21/FB
	 */
	
	private void init_and_log_Request(String func, String requestPath)
	
	{
	String
		realPath = servletContext.getRealPath(fnameProps);

	// load GlemmServices properties and log them (if last param = true):
	props = Utils.loadProperties(servletContext.getResourceAsStream(fnameProps), fnameProps, realPath, false);
	
	set_ConstantsFromProperties(props);
	
	// Set Standard File output streams for GlemmServices:
	// Utils.setStdFileStreams(fnameErr, fnameOut);
	
	// open fout and ferr as replacement for stdout/stderr for use in GlemmServices.
	// do not use or redirect System.out and System.err, as they will be redirected
	// by other Services in the same glassfish JVM, e.g. by the WformServices.
	fout = Utils.setLogStream(fnameOut);
	ferr = Utils.setLogStream(fnameErr);
	
	log_Request(requestPath, func);

	} // init_and_log_Request

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 * GET: /wformServices/status/
	 * 
	 * returns status of Wordform Services in HTML.
	 * 
	 * Parameters:
	 *  -
	 * Notes:
	 * - 
	 * 12.06.20/FB 
	 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	@Path("status")
	@GET
	@Produces("text/html")
	
	public String doGET_status()
	
	{
	final String func = "doGET_status";
	String
		status;

	init_and_log_Request(func, "/wformServices/status/");
	
	status = String.format(	"<html>\n" +
						   	" <h3>Wordform Services</h3>\n" +
						   	" <p><b>Version</b>     : %s</p>" +
							" <p><b>Date   </b>     : %s</p>" +
							" <p><b>status</b>      : ready.</p>" +
							" <p><b>Request time</b>: %s</p>" +
							" <p><b>properties</b>: %s </p>" +
							"</html>",
							version,
							versionDate,
							LocalDateTime.now().toString().replace("T", " "),
							servletContext.getResourceAsStream(fnameProps) == null ? "not found" :
								props.isEmpty() ? "empty" : "loaded");

	fout.printf("Debug: %s: done.\n", func);

	return status;
		
	} // doGET_status
	
	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 * GET: /wformServices/plugin.html
	 * 
	 * returns plugin.html file for registration in Kalamar.
	 * 
	 * Parameters:
	 *  -
	 * Notes:
	 * - 
	 * 29.06.20/FB 
	 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	@Path("plugin.html")
	@GET
	@Produces("text/html")
	
	public String doGET_pluginHTML(@Context ServletContext servletContext)
	
	{
	final String func = "doGET_pluginHTML";
	String
		pluginHTML;
	String
		webAppRoot;

	init_and_log_Request(func, "/wformServices/plugin.html");
	
	webAppRoot = servletContext.getRealPath("/");
	
	fout.printf("Debug: %s: Root Path of webApp = '%s'.\n",  func, webAppRoot);
	fout.printf("Debug: %s: loading = '%s'.\n",  func, webAppRoot + filename_pluginHTML);
	
	try {
		pluginHTML = Utils.readFromInputStream(webAppRoot + filename_pluginHTML);
		fout.printf("Debug: %s: loading plugin.html: OK'.\n",  func);
		fout.printf("Debug: %s: plugin.html = '%s'.\n",  func, pluginHTML);
		} 
	catch (IOException e) {
		e.printStackTrace();
		pluginHTML = String.format("<html>\n<body><h3>wformServices Error: cannot load plugin.html!</h3></body></html>\n");
		fout.printf("Debug: %s: loading plugin.html: failed!\n",  func);
		}
	
	fout.printf("Debug: %s: done.\n", func);
	return pluginHTML;
		
	} // doGET_pluginHTML
	
	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 * POST: /wformServices/query/: application/json.
	 * submits a Koral Query which might contain 1 or more lemma subqueries [UTF-8].
	 * Each subquery is searched in the Glemm DB and the resulting wordform list rewrites
	 * the original subquery.
	 * Parameters:
	 *  query    : content of POST body: full KoralQuery in Json.
	 * Notes:
	 * - HTTP POST Body can be read once only. Probably Jersey reads it once and
	 *   passes it to the POST method, i.e. String query.
	 * - if we try to read the POST Body by either HttpServletRequest.getReader()
	 *   or HttpServletRequest.getInputStream(), we get an exception, as this 
	 *   would be the 2nd access to the body data.
	 * 17.02.20/FB 
	 * 20.02.20/FB Body Data access by parameter query.
	 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
	@Path("query") 
	@POST 
	@Consumes("application/json") 
	@Produces("application/json")
	 
	public String doPOST_handleWordformQueries(String query) 
	 
	{
	final String
		func = "doPOST_handleWordformQueries";
	String
		newQuery = "";
	
	init_and_log_Request(func, "/wformServices/query/");
	
	try	{
		newQuery = WordformRewrite.traverse_and_rewrite_KoralQuery(query, props, fout, ferr);
		if( newQuery == null )
			{
			fout.printf("Debug: %s: ending time: %s.\n", func, LocalDateTime.now().toString().replace("T",  " "));
			fout.printf("Debug: %s: nothing to rewrite.\n", func);
			return query; // nothing to rewrite, return original query.
			}
		}
	catch( Exception e)
		{
		e.printStackTrace();
		fout.printf("Debug: %s: ending time: %s.\n", func, LocalDateTime.now().toString().replace("T",  " "));
		fout.printf("Debug: %s: Exception occured: returning unrewritten query!\n", func);
		return query; // error occured, but returning original, unrewritten query.
		}
	
	fout.printf("Debug: %s: ending time: %s.\n", func, LocalDateTime.now().toString().replace("T",  " "));
	fout.printf("Debug: %s: returning rewritten query: size=%d chars.\n", func, newQuery.length());
	
	return newQuery;
	
	} // doPOST_handleWordformQueries

}
